﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models;
using ClientServer.Models.CustomProjects;
using ClientServer.Models.Exercises;
using ClientServer.Models.Exercises.Release;
using ClientServer.Models.Exercises.Solution;
using ClientServer.Models.Exercises.Tests;
using ClientServer.Models.Files;
using ClientServer.Models.Users;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;

namespace ClientServer.Controllers.Core
{
    [Route(Constants.InitControllerAbsolutePath)]
    public class Init : Controller
    {
        private readonly YapexDbContext _context;


        public Init(YapexDbContext context)
        {
            _context = context;
        }

        [HttpGet]
        public void Get()
        {
            //check if at least 1 user exists... 

            var firstUser = _context.Users.FirstOrDefault();

            if (firstUser != null)
            {
                //at least on user exits --> do not insert basic data twice
                return;
            }
            else
            {
                //else we need to create some basic data
                // e.g. first user, test types plangs, langs, ... 
            }

            Console.WriteLine("Creating first login (id: j, pw: j), CHANGE ADMIN CREDENTIALS AFTER LOGIN");

            //TODO remove

            //var blog = new Blog()
            //{
            //    Name = "BLOG",
            //    Posts = new List<Post>()
            //};

            //var post1 = new Post()
            //{
            //    Title = "t1",
            //};

            //var post2 = new Post()
            //{
            //    Title = "t2"
            //};

            //_context.Blogs.Add(blog);
            //_context.SaveChanges();

            //blog.Posts.Add(post1);
            //blog.Posts.Add(post2);
            //blog.MainPost = post1;


            //_context.SaveChanges();


            //one admin
            //one default group with role for all new users (students)
            //one default role for group creators

            #region set up first group independent 

            var adminRolePermission = new SystemRolePermission()
            {
                CanChangeRoles = true,
                CanCreateGroups = true,
                CanCreateRoles = true,
                CanDeleteGroups = true,
                CanDeleteRoles = true,
                CanChangeOtherUsersSystemRole = true,
                CanManageNewUsers = true,
                CanDeleteActivatedUsers = true,
                CanChangeSystemSettings = true,
                CanChangeUserData = true,
                CanManageTags = true,
                CanViewDashboard = true,
            };

            //the first system role cannot be changed/deleted!
            var adminSystemRole = new SystemRole()
            {
                DisplayName = Constants.DefaultSystemRoleDisplayName,
                Email = "admins@admins.de",
                SystemRolePermission = adminRolePermission
            };

            #endregion

            var adminPw = PasswordHelper.GetHash(PasswordHelper.HashLikeFrontend("j"));

            var admin = new User()
            {
                FirstName = "Admin",
                LastName = "Admin",
                Email = "admin@example.de",
                IsActivated = true,
                Token = "j", //admin //ahbxf
                Password = adminPw,
                SystemRole = adminSystemRole,
            };

            #region create first real group default group for all new users (students)

            var studentGroupRolePermission = new GroupRolePermission(false, false, false, false, false, false, false,
                false, false, false, false, false
            );

            var studentGrouRole = new GroupRole()
            {
                Id = InitialSystemSettings._DefaultGroupRoleId, //i don't think we can force this id but we can try...
                DisplayName = Constants.DefaultGroupRoleDisplayName,
                GroupRolePermission = studentGroupRolePermission
            };

            _context.GroupRoles.Add(studentGrouRole);

            //default group for all new users
            var studentGroup = new UserGroup()
            {
                Id = InitialSystemSettings._DefaultUserGroupId, //i don't think we can force this id but we can try...
                Email = "default@example.com",
                DisplayName = Constants.DefaultGroupDisplayName,
            };

            _context.UserGroups.Add(studentGroup);

            #endregion

            #region create creator (of groups) role and permission

            var creatorGroupRolePermission = new GroupRolePermission(true, true, true, true, true, true, true,
                true, true, true, true, true
            );

            var creatorGroupRole = new GroupRole()
            {
                Id = InitialSystemSettings
                    ._DefaultGroupCreatorGroupRoleId, //i don't think we can force this id but we can try...
                DisplayName = Constants.GroupCreatorRoleDisplayName,
                GroupRolePermission = creatorGroupRolePermission,
            };

            //the admin is the "creator" of the students group 
            var firstUserConnection = new UserWithUserGroup()
            {
                User = admin,
                UserGroup = studentGroup,
                GroupRole = creatorGroupRole,
            };

            _context.UserWithUserGroups.Add(firstUserConnection);

            #endregion

            #region add some programming langs

            //var csLang = new PLang()
            //{
            //    DisplayName = "C#",
            //    InternalName = "csharp"
            //};

            //_context.PLangs.Add(csLang);


            var javaLang = new PLang()
            {
                DisplayName = "Java",
                InternalName = "java",
                FileExtensionsWithDot = ".java",
                EditorHighlightModeName = "java"
            };
            _context.PLangs.Add(javaLang);


            //var pyLang = new PLang()
            //{
            //    DisplayName = "Python",
            //    InternalName = "python"
            //};

            //_context.PLangs.Add(pyLang);

            _context.Langs.Add(new Lang()
            {
                Language = "English",
                LangShortcut = "en"
            });

            _context.Langs.Add(new Lang()
            {
                Language = "Deutsch",
                LangShortcut = "de"
            });

            #endregion

            var blackBoxTest = new TestType()
            {
                DisplayName = "Black-Box-Test",
                InternalName = Constants.BlackBoxTestInternalName
            };
            var syntaxTest = new TestType()
            {
                DisplayName = "Syntaxtest",
                InternalName = Constants.CompileTestInternalName
            };
            var regexTest = new TestType()
            {
                DisplayName = "Regex-Test",
                InternalName = Constants.RegexTestInternalName
            };
            var compareFilesTest = new TestType()
            {
                DisplayName = "Dateienvergleich-Test",
                InternalName = Constants.CompareFilesTestInternalName
            };

            _context.TestTypes.Add(blackBoxTest);
            _context.TestTypes.Add(syntaxTest);
            _context.TestTypes.Add(regexTest);
            _context.TestTypes.Add(compareFilesTest);

            var systemSettings = new SystemSetting()
            {
                DefaultGroupCreatorGroupRole = creatorGroupRole,
                DefaultGroupRole = studentGrouRole,
                DefaultUserGroup = studentGroup,
                MaxCustomProjectsPerUser = InitialSystemSettings.MaxCustomProjectsPerUser,
                MaxCustomTestsPerParticipation = InitialSystemSettings.MaxCustomTestsPerParticipation,
                CurrentServerMessage = InitialSystemSettings.CurrentServerMessage,
                TestServerStatsUrl = InitialSystemSettings.TestServerStatsUrl,
                SubmitTestServerUrl = InitialSystemSettings.SubmitTestServerUrl,
                TestServerUrl = InitialSystemSettings.TestServerUrl,
                TestServerConfigUiUrl = InitialSystemSettings.TestServerConfigUiUrl,
                SubmitTestServerTimeoutInMs = InitialSystemSettings.SubmitTestServerTimeoutInMs,
                TestServerTimeoutInMs = InitialSystemSettings.TestServerTimeoutInMs,
                MaxNumberOfTestsWithOneRequest = InitialSystemSettings.MaxNumberOfTestsWithOneRequest,
                MaxNumberOfTestsWithOneRequestSubmitTestServer =
                    InitialSystemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer,
                JustRunProgramTimeoutInMs = InitialSystemSettings.JustRunProgramTimeoutInMs,
                JustRunProgramCompileTimeoutInMs = InitialSystemSettings.JustRunProgramCompileTimeoutInMs,
                JustRunProgramMemoryLimitInKb = InitialSystemSettings.JustRunProgramMemoryLimitInKb,
                JustRunProgramMaxDiskSpaceInKb = InitialSystemSettings.JustRunProgramMaxDiskSpaceInKb,
                CustomProjectTestTimeoutInMs = InitialSystemSettings.CustomProjectTestTimeoutInMs,
                CustomProjectTestCompileTimeoutInMs = InitialSystemSettings.CustomProjectTestCompileTimeoutInMs,
                CustomProjectTestMemoryLimitInKb = InitialSystemSettings.CustomProjectTestMemoryLimitInKb,
                CustomProjectTestMaxDiskSpaceInKb = InitialSystemSettings.CustomProjectTestMaxDiskSpaceInKb,
            };


            _context.SystemSettings.Add(systemSettings);

            #region test add inactive users

            //for (int i = 0; i < 5; i++)
            //{
            //    var newUser = new User()
            //    {
            //        Email = "example@example.de",
            //        IsActivated = false,
            //        FirstName = "Test " + (i + 1),
            //        LastName = "Tester" + (i + 1),
            //        Password = PasswordHelper.GetHash((i + 1) + ""),
            //        Token = "j" + (i + 1),
            //        UserWithUserGroups = new List<UserWithUserGroup>()
            //        {
            //            new UserWithUserGroup()
            //            {
            //                UserGroup = systemSettings.DefaultUserGroup,
            //                GroupRole = systemSettings.DefaultGroupRole
            //            }
            //        }
            //    };
            //    _context.Users.Add(newUser);
            //}

            #endregion

            int changed = _context.SaveChanges();


            //create example exercise

            Response.WriteAsync("OK -" + changed);
        }


        [HttpGet("migrate/addfks")]
        public void MigrateMarkdownFiles()
        {
            var systemRoles = _context.SystemRoles
                .Include(p => p.SystemRolePermission)
                .ToList();

            foreach (var systemRole in systemRoles)
            {
                systemRole.SystemRolePermission.SystemRoleId = systemRole.Id;
            }


            var groupRoles = _context.GroupRoles
                    .Include(p => p.GroupRolePermission)
                    .ToList()
                ;

            foreach (var groupRole in groupRoles)
            {
                groupRole.GroupRolePermission.GroupRoleId = groupRole.Id;
            }


            var userSettings = _context.UserSettingses
                .Include(p => p.CodeEditorSetting)
                .ToList();

            foreach (var userSetting in userSettings)
            {
                userSetting.CodeEditorSetting.UserSettingId = userSetting.Id;
            }


            var solutions = _context.Solutions
                .Include(p => p.Assessment)
                .ToList();

            solutions = solutions.Where(p => p.Assessment != null).ToList();

            foreach (var solution in solutions)
            {
                solution.Assessment.PLangId = solution.PLangId;
                solution.Assessment.UserId = solution.UserId;
                solution.Assessment.ExerciseReleaseId = solution.ExerciseReleaseId;
            }


            var allTestSettings = _context.TestCaseSettingses.ToList();


            var allValidTestSettingsIds = new List<int>();

            var exTests = _context.Tests
                    .Include(p => p.TestSettings)
                    .ToList()
                ;

            foreach (var exTest in exTests)
            {
                exTest.TestSettings.TestId = exTest.Id;
                allValidTestSettingsIds.Add(exTest.TestSettings.Id);
            }

            //we need to delete the old test settings... BUT it has on delete cascade
            //so first migrate then drop the constraint and then remove the old test settings...
            var customProjectTests = _context.CustomProjectTests
                .Include(p => p.CustomProjectTestSettings)
                .ToList();


            foreach (var customProjectTest in customProjectTests)
            {
                var customProjectTestSettings = new CustomProjectTestSettings()
                {
                    Id = customProjectTest.CustomProjectTestSettings.Id,
                    CompileTimeoutInMs = customProjectTest.CustomProjectTestSettings.CompileTimeoutInMs,
                    CompilerOptions = customProjectTest.CustomProjectTestSettings.CompilerOptions,
                    TimeoutInMs = customProjectTest.CustomProjectTestSettings.TimeoutInMs,
                    MemoryLimitInKb = customProjectTest.CustomProjectTestSettings.MemoryLimitInKb,
                    MaxDiskSpaceInKb = customProjectTest.CustomProjectTestSettings.MaxDiskSpaceInKb,
                    CustomProjectTestId = customProjectTest.Id
                };

                _context.CustomProjectTestSettings.Add(customProjectTestSettings);
                allValidTestSettingsIds.Add(customProjectTest.CustomProjectTestSettings.Id);
            }


            var count1 = exTests.Count + customProjectTests.Count;
            var removedCount = 0;


            //because we have the delete cascade on the wrong side... we never delete test settings...
            foreach (var allTestSetting in allTestSettings)
            {
                if (allValidTestSettingsIds.Contains(allTestSetting.Id))
                {
                    //valid
                }
                else
                {
                    _context.TestCaseSettingses.Remove(allTestSetting);
                    removedCount++;
                }
            }

            var newCount2 = allTestSettings.Count - removedCount;
            if (newCount2 != count1)
            {
                Response.WriteAsync("wrong");
                return;
            }

            int rows;
            try
            {
                rows = _context.SaveChanges();
            }
            catch (Exception e)
            {
                Console.WriteLine(e);
                throw;
            }


            Response.WriteAsync("OK -" + rows);
        }

        //remove test case settings fk on delete cascade... because when we add the new CustomProjectTestSettings we want to delete the old

        //remove the old test settings where we created a new CustomProjectTestSettings for (testId is 0) 

        [HttpGet("migrate/step2")]
        public void MigrateMarkdownFiles2()
        {
            var allTestSettings = _context.TestCaseSettingses
                .Where(p => p.TestId == 0)
                .ToList();
            
            _context.TestCaseSettingses.RemoveRange(allTestSettings);

            
            

            int rows;
            try
            {
                var notDeletedAssessments = _context.SolutionAssessment
                    .Where(p => p.UserId == 0 && p.ExerciseReleaseId == 0 && p.PLangId == 0)
                    .ToList();
            
            
                _context.SolutionAssessment.RemoveRange(notDeletedAssessments);
                rows = _context.SaveChanges();
            }
            catch (Exception e)
            {
                Console.WriteLine(e);
                throw;
            }

            Response.WriteAsync("OK -" + rows);
        }
    }
}
